import 'package:apex_socket/cipher/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:socket_io_client/socket_io_client.dart' as IO;
import 'package:logging/logging.dart';

export 'package:socket_io_client/socket_io_client.dart';

typedef dynamic EventHandler<T>(T data);

class ApexSocket extends ChangeNotifier {
  final log = Logger('ApexSocket');

  IO.Socket socket;

  int _reconnectDelay = 2;

  /// Check if socket object is created and connected as well :)
  bool get isConnected => socket != null && socket.connected;

  /// Reconnection delay in seconds
  int get reconnectionDelay => _reconnectDelay;

  final int port;
  final String url;

  /// Just works on encrypt = true
  final String prefixData;
  final String namespace;
  final IO.OptionBuilder options;
  final bool encrypt;
  final String _secretKey;
  final String _publicKey;

  ApexSocket(this.url, this.port, this.options,
      {this.prefixData,
      this.namespace,
      this.encrypt = false,
      String secretKey,
      String publicKey})
      : _secretKey = secretKey,
        _publicKey = publicKey,
        assert(
            (encrypt && secretKey != null && publicKey != null) || (!encrypt)),
        assert(url != null && Uri.parse(url).isAbsolute,
            '$url must be a valid url.'),
        assert(port == null || (port >= -1 && port <= 65535),
            '$port must be a number between -1 and 65535. or null.'),
        assert(options != null, 'Options must not be null');

  void connect() {
    if (this.socket != null) this.socket.connect();
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (this.socket != null) this.socket.destroy();
      IO.Socket socket = IO.io(
          '$url${port != null ? ':$port' : ''}${namespace != null ? '/$namespace' : ''}',
          options
              .setReconnectionDelay(_reconnectDelay)
              .setTransports(['websocket']).build());

      socket.onConnect((data) {
        _reconnectDelay = 2;
        notifyListeners();
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        notifyListeners();
        if (onDisconnect != null) return onDisconnect(data);
      });

      socket.onReconnect(onReconnect);

      socket.onReconnecting((data) {
        if (_reconnectDelay < 64) {
          _reconnectDelay *= 2;
        } else {
          _reconnectDelay = 64;
        }
        socket.io.options =
            options.setReconnectionDelay(_reconnectDelay).build();
        notifyListeners();
      });

      this.socket = socket;
    }

    if (!this.socket.connected)
      this.socket.connect();
    else
      log.info('already connected');
  }

  void emit(String event, [String data]) {
    if (encrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      data = (prefixData ?? '').replaceAll(':', '') + ':' + crypto.encrypt(data);
    }
    this.socket.emit(event, data);
  }

  void emitWithAck(String event, String data, {Function(String data) ack}) {
    Crypto crypto;
    if (encrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      data = (prefixData ?? '').replaceAll(':', '') + ':' + crypto.encrypt(data);
    }
    this.socket.emitWithAck(event, data, ack: (m) {
      if (encrypt)
        ack(crypto.decrypt(m.toString()));
      else
        ack(m.toString());
    });
  }
}
